/*
 * Use GL_ARB_fragment_shader and GL_ARB_vertex_shader to implement
 * simple per-pixel lighting.
 *
 * Ian Romanick
 * 20 January 2007
 * Updated to use SDL instead of GLUT.
 * 
 * Michal Krol
 * 20 February 2006
 *
 * Based on the original demo by:
 * Brian Paul
 * 17 April 2003
 */

#ifdef WIN32
#include <windows.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <GL/gl.h>
#include <GL/glext.h>

#include <SDL/SDL.h>

static GLfloat diffuse[4] = { 0.5f, 0.5f, 1.0f, 1.0f };
static GLfloat specular[4] = { 0.8f, 0.8f, 0.8f, 1.0f };
static GLfloat lightPos[4] = { 0.0f, 10.0f, 20.0f, 1.0f };
static GLfloat delta = 1.0f;

static GLhandleARB fragShader;
static GLhandleARB vertShader;
static GLhandleARB program;

static GLint uLightPos;
static GLint uDiffuse;
static GLint uSpecular;

static GLboolean anim = GL_TRUE;
static GLboolean wire = GL_FALSE;
static GLboolean pixelLight = GL_TRUE;
static bool done = false;

static SDL_Surface *my_surf = NULL;

static GLint t0 = 0;
static GLint frames = 0;

static GLfloat xRot = 0.0f, yRot = 0.0f;

static PFNGLCREATESHADEROBJECTARBPROC glCreateShaderObjectARB = NULL;
static PFNGLSHADERSOURCEARBPROC glShaderSourceARB = NULL;
static PFNGLCOMPILESHADERARBPROC glCompileShaderARB = NULL;
static PFNGLCREATEPROGRAMOBJECTARBPROC glCreateProgramObjectARB = NULL;
static PFNGLATTACHOBJECTARBPROC glAttachObjectARB = NULL;
static PFNGLLINKPROGRAMARBPROC glLinkProgramARB = NULL;
static PFNGLUSEPROGRAMOBJECTARBPROC glUseProgramObjectARB = NULL;
static PFNGLGETUNIFORMLOCATIONARBPROC glGetUniformLocationARB = NULL;
static PFNGLUNIFORM3FVARBPROC glUniform3fvARB = NULL;
static PFNGLUNIFORM3FVARBPROC glUniform4fvARB = NULL;

static bool extension_supported(const char *ext_name)
{
    return true;
}

void
draw_solid_sphere(double radius, unsigned slices, unsigned stacks)
{
    const float lat_inc = M_PI / (float) stacks;
    const float lon_inc = (2.0 * M_PI) / (float) slices;

    float lat_angle = 0.0;

    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();

    glEnable(GL_NORMALIZE);
    glScalef(radius, radius, radius);

    for (unsigned i = 0; i < stacks; i++) {
	const float y = cos(lat_angle);
	const float y_next = cos(lat_angle + lat_inc);
	const float r = sin(lat_angle);
	const float r_next = sin(lat_angle + lat_inc);
	const float v = (float) i / (float) stacks;
	const float v_next = (float) (i + 1) / (float) stacks;

	float lon_angle = 0.0;

	glBegin(GL_QUAD_STRIP);
	for (unsigned j = 0; j < slices; j++) {
	    const float x = cos(lon_angle);
	    const float z = sin(lon_angle);
	    const float u = (float) j / (float) slices;

	    glNormal3f(x * r, y, z * r);
	    glVertex3f(x * r, y, z * r);

	    glNormal3f(x * r_next, y_next, z * r_next);
	    glVertex3f(x * r_next, y_next, z * r_next);

	    lon_angle += lon_inc;
	}

	const float x = cos(lon_angle);
	const float z = sin(lon_angle);

	glNormal3f(x * r, y, z * r);
	glVertex3f(x * r, y, z * r);

	glNormal3f(x * r_next, y_next, z * r_next);
	glVertex3f(x * r_next, y_next, z * r_next);

	glEnd();

	lat_angle += lat_inc;
    }

    glDisable(GL_NORMALIZE);
    glPopMatrix();
}

static void normalize (GLfloat *dst, const GLfloat *src)
{
   GLfloat len = sqrtf (src[0] * src[0] + src[1] * src[1] + src[2] * src[2]);
   dst[0] = src[0] / len;
   dst[1] = src[1] / len;
   dst[2] = src[2] / len;
}

static void Redisplay (void)
{
    glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    if (pixelLight) {
	GLfloat vec[3];

	glUseProgramObjectARB (program);
	normalize (vec, lightPos);
	glUniform3fvARB (uLightPos, 1, vec);
	glDisable(GL_LIGHTING);
    }
    else {
	glUseProgramObjectARB (0);
	glLightfv (GL_LIGHT0, GL_POSITION, lightPos);
	glEnable(GL_LIGHTING);
    }

    glPushMatrix ();
    glRotatef (xRot, 1.0f, 0.0f, 0.0f);
    glRotatef (yRot, 0.0f, 1.0f, 0.0f);
    draw_solid_sphere (2.0, 10, 5);
    glPopMatrix ();

    SDL_GL_SwapBuffers();
    frames++;

    if (anim) {
	GLuint t = SDL_GetTicks();
	if (t - t0 >= 5000) {
	    GLfloat seconds = (GLfloat) (t - t0) / 1000.0f;
	    GLfloat fps = frames / seconds;
	    printf ("%d frames in %6.3f seconds = %6.3f FPS\n", frames, seconds, fps);
	    t0 = t;
	    frames = 0;
	}
    }
}

static void Idle (void)
{
    lightPos[0] += delta;
    if (lightPos[0] > 25.0f || lightPos[0] < -25.0f)
	delta = -delta;
}

static void Reshape (int width, int height)
{
    glViewport (0, 0, width, height);
    glMatrixMode (GL_PROJECTION);
    glLoadIdentity ();
    glFrustum (-1.0, 1.0, -1.0, 1.0, 5.0, 25.0);
    glMatrixMode (GL_MODELVIEW);
    glLoadIdentity ();
    glTranslatef (0.0f, 0.0f, -15.0f);
}

static void Key(SDLKey sym)
{
    const GLfloat step = 3.0f;

    switch(sym) {
    case SDLK_ESCAPE:
	done = true;
	break;
                    
    case 'f':
	SDL_WM_ToggleFullScreen(my_surf);
	break;
    case ' ':
    case 'a':
	anim = !anim;
	break;
    case 'x':
	lightPos[0] -= 1.0f;
	break;
    case 'X':
	lightPos[0] += 1.0f;
	break;
    case 'w':
	wire = !wire;
	if (wire)
	    glPolygonMode (GL_FRONT_AND_BACK, GL_LINE);
	else
	    glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);
	break;
    case 'p':
	pixelLight = !pixelLight;
	if (pixelLight)
	    printf ("Per-pixel lighting\n");
	else
	    printf ("Conventional lighting\n");
	break;
    case SDLK_UP:
	xRot -= step;
	break;
    case SDLK_DOWN:
	xRot += step;
	break;
    case SDLK_LEFT:
	yRot -= step;
	break;
    case SDLK_RIGHT:
	yRot += step;
	break;
    default:
	break;
    }
}

static void Init (void)
{
   static const char *fragShaderText =
      "uniform vec3 lightPos;\n"
      "uniform vec4 diffuse;\n"
      "uniform vec4 specular;\n"
      "varying vec3 normal;\n"
      "void main () {\n"
      "   // Compute dot product of light direction and normal vector\n"
      "   float dotProd = max (dot (lightPos, normalize (normal)), 0.0);\n"
      "   // Compute diffuse and specular contributions\n"
      "   gl_FragColor = diffuse * dotProd + specular * pow (dotProd, 20.0);\n"
      "}\n"
   ;
   static const char *vertShaderText =
      "varying vec3 normal;\n"
      "void main () {\n"
      "   gl_Position = gl_ModelViewProjectionMatrix * gl_Vertex;\n"
      "   normal = gl_NormalMatrix * gl_Normal;\n"
      "}\n"
   ;

    if (!extension_supported("GL_ARB_fragment_shader")) {
	printf ("Sorry, this demo requires GL_ARB_fragment_shader\n");
	exit(1);
    }
    if (!extension_supported("GL_ARB_shader_objects")) {
	printf ("Sorry, this demo requires GL_ARB_shader_objects\n");
	exit(1);
    }
    if (!extension_supported("GL_ARB_shading_language_100")) {
	printf ("Sorry, this demo requires GL_ARB_shading_language_100\n");
	exit(1);
    }
    if (!extension_supported("GL_ARB_vertex_shader")) {
	printf ("Sorry, this demo requires GL_ARB_vertex_shader\n");
	exit(1);
    }

    glCreateShaderObjectARB = (PFNGLCREATESHADEROBJECTARBPROC) SDL_GL_GetProcAddress ("glCreateShaderObjectARB");
    glShaderSourceARB = (PFNGLSHADERSOURCEARBPROC) SDL_GL_GetProcAddress ("glShaderSourceARB");
    glCompileShaderARB = (PFNGLCOMPILESHADERARBPROC) SDL_GL_GetProcAddress ("glCompileShaderARB");
    glCreateProgramObjectARB = (PFNGLCREATEPROGRAMOBJECTARBPROC) SDL_GL_GetProcAddress ("glCreateProgramObjectARB");
    glAttachObjectARB = (PFNGLATTACHOBJECTARBPROC) SDL_GL_GetProcAddress ("glAttachObjectARB");
    glLinkProgramARB = (PFNGLLINKPROGRAMARBPROC) SDL_GL_GetProcAddress ("glLinkProgramARB");
    glUseProgramObjectARB = (PFNGLUSEPROGRAMOBJECTARBPROC) SDL_GL_GetProcAddress ("glUseProgramObjectARB");
    glGetUniformLocationARB = (PFNGLGETUNIFORMLOCATIONARBPROC) SDL_GL_GetProcAddress ("glGetUniformLocationARB");
    glUniform3fvARB = (PFNGLUNIFORM3FVARBPROC) SDL_GL_GetProcAddress ("glUniform3fvARB");
    glUniform4fvARB = (PFNGLUNIFORM3FVARBPROC) SDL_GL_GetProcAddress ("glUniform4fvARB");

    fragShader = glCreateShaderObjectARB (GL_FRAGMENT_SHADER_ARB);
    glShaderSourceARB (fragShader, 1, &fragShaderText, NULL);
    glCompileShaderARB (fragShader);

    vertShader = glCreateShaderObjectARB (GL_VERTEX_SHADER_ARB);
    glShaderSourceARB (vertShader, 1, &vertShaderText, NULL);
    glCompileShaderARB (vertShader);

    program = glCreateProgramObjectARB ();
    glAttachObjectARB (program, fragShader);
    glAttachObjectARB (program, vertShader);
    glLinkProgramARB (program);
    glUseProgramObjectARB (program);

    uLightPos = glGetUniformLocationARB (program, "lightPos");
    uDiffuse = glGetUniformLocationARB (program, "diffuse");
    uSpecular = glGetUniformLocationARB (program, "specular");

    glUniform4fvARB (uDiffuse, 1, diffuse);
    glUniform4fvARB (uSpecular, 1, specular);

    glClearColor (0.3f, 0.3f, 0.3f, 0.0f);
    glEnable (GL_DEPTH_TEST);
    glEnable (GL_LIGHT0);
    glEnable (GL_LIGHTING);
    glMaterialfv (GL_FRONT_AND_BACK, GL_DIFFUSE, diffuse);
    glMaterialfv (GL_FRONT_AND_BACK, GL_SPECULAR, specular);
    glMaterialf (GL_FRONT_AND_BACK, GL_SHININESS, 20.0f);

    printf ("GL_RENDERER = %s\n", (const char *) glGetString (GL_RENDERER));
    printf ("Press p to toggle between per-pixel and per-vertex lighting\n");
}

Uint32 timer_callback(Uint32 interval, void *not_used)
{
    SDL_Event e;

    (void) not_used;

    e.type = SDL_USEREVENT;
    e.user.code = 0;
    e.user.data1 = NULL;
    e.user.data2 = NULL;
    SDL_PushEvent(& e);

    return interval;
}


int main (int argc, char *argv[])
{
    if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER) < 0) {
	exit(1);
    }
    
    my_surf = SDL_SetVideoMode(800, 600, 0, SDL_OPENGL);
    if (my_surf == NULL) {
	SDL_Quit();
	exit(1);
    }

    Reshape(800, 600);
    Init();

    SDL_TimerID timer_id = SDL_AddTimer(10, timer_callback, NULL);

    while (!done) {
        SDL_Event event;

        SDL_WaitEvent(&event);
	do {
            switch (event.type) {
            case SDL_VIDEORESIZE:
                Reshape(event.resize.w, event.resize.h);
                break;

            case SDL_QUIT:
                done = true;
                break;
                
            case SDL_KEYDOWN:
		Key(event.key.keysym.sym);
                break;

            default:
                break;
            }
        } while (SDL_PollEvent(&event));


	Idle();
	Redisplay();
    }

    SDL_RemoveTimer(timer_id);

    SDL_Quit();
    return 0;
}

